//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//
//  crccalc.h (libcoyote)
//
//  Defines tools for calculating 32-bit cyclic-redundancy (CRC)
//  values from arrays of 8-bit byte data.
//---------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 2000, 2002, 2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//
//  This computer program a_source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this a_source code, or portions hereof, for any purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this a_source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original a_source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         a_source or altered a_source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this a_source code for entertainment, education, or decoration. If
//  you use this a_source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

#if !defined(LIBCOYOTE_CRCCALC_H)
#define LIBCOYOTE_CRCCALC_H

namespace libcoyote
{
    //! 32-bit unsigned value for holding CRC value
    typedef unsigned long crc32_t;
    
    //! Calculate 32-bit CRC values from byte data
    /*!
        Calculates a 32-bit CRC value using the algorithm specified by 
        ISO 3309, and as used by the PNG graphics specification.
        \version 1.1.0
        \date    15 June 2002
    */
    class crc_calculator
    {
        // internal class for preclaculated table
        private:
            class crc_precalc  
            {
            public:
                // constructor
                crc_precalc();
            
                // return element of table
                crc32_t operator [] (int n) const
                {
                    return m_table[n];
                }
            
            protected:
                // table of precalculated values
                crc32_t m_table[256];
            };
        
        public:
            //! Constrctor
            /*!
                Constructs a new CRC value with a base value of 0xFFFFFFFF.
            */
            crc_calculator();
        
            //! Update CRC for a given data set
            /*!
                Updates the crc value from the bytes provided in <i>data</i>.
                Each time this function is called for a given crc_calculator,
                the 32-bit CRC value changes to reflect the new bytes.
                \param a_data - Array of bytes to be "added" to the CRC value
                \param a_length - Number of bytes in the <i>data</i> array
            */  
            void update(const unsigned char * a_data, size_t a_length);
        
            //! Get CRC value
            /*!
                Returns the current 32-bit CRC value. This value will change
                with subsequent calls to update.
                \return Current 32-bit CRC value
            */
            crc32_t get_crc32()
            {
                return m_crc ^ 0xFFFFFFFFL;
            }
        
        private:
            // precacluated table
            static crc_precalc s_table;
    
            // computed CRC value
            crc32_t m_crc;
    };
    
} // end namespace libcoyote

#endif
